/************************************************************************
  XBox media server v0.30.4

  xbms is a *nix streaming server for xbox mediaplayer. 
  It provides access to shared directory on a unix box right from the 
  xboxmediaplayer2  (www.xboxmediaplayer.de).

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Ugly code by J-F Mammet <jeff@bidou.info>
  Not ugly code by BigUglyFool
  Network protocol by Runtime
  XBPlayer by Runtime & d7o3g4q
  Header cleanup by Stefan Mansby <stefan@nis.nu>
  IP binding and some buffer fixing by tric
  enabled using "user" to run as and added group by tric (group might not work)
  first Player2 support (by tric)
  added forking and mmaped mem, i know its not clean, but hey it works (tric)
  ok, fixed some warnings and i'll use CHANGELOG from now on. (tric)

  TODO:
  - test player2 support and add iso/bin support
  - cleanup code, find overflow and security issues
  
*************************************************************************/

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/wait.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <dirent.h>   
#include <time.h>
#include <fcntl.h>
#include <stdlib.h>
#include <pwd.h>
#include <grp.h>
#include <errno.h>
#include <sys/mman.h>

#define SUCCESS 0
#define ERROR   1


#define CONFIG "/etc/xbms.conf"
#define PIDFILE "/var/run/xbms.pid"
#define C_SERVER_PORT 1400
#define MAX_MSG_LENGTH 4096
#define MAX_NAME_LENGTH 4096 
#define MAX_CMD_LENGTH 4
#define LISTEN_ADDR "0.0.0.0"
/* user and group will only be used if running as root
 * or if you pass it by cmd line						*/
#define USER_NAME "nobody"
#define GROUP_NAME "nogroup"

/* This should be the only thing to configure for the moment */
#define MEDIA_FILES_PATH "/"


char *LOG_FILE;
static int RUNNING=1;

extern char *ctime(const time_t *time_tzgr);
extern time_t time(time_t *time_tzgr);

int usage(char *prog)
{
  printf("Usage:");
  printf("%s [options]\n\n", prog);
  printf("\t-i [ip-address]");
  printf("\t-d [media dir]\n");
  printf("\t-p [port]\n");
  printf("\t-u [username to run as]\n");
  printf("\t-g [group to run as]\n");
  printf("\t-D [level] debug mode (lvl 2 VERY verbose)\n");
  printf("\t-c [config file]\n");
  printf("\t-l [logfile]\n");
  printf("\t-h help\n");
  exit(1);
}

/* This function does not make any safety check, beware ! */
int substr(char *buf, char **dest, int s, int l,int e)
{
  //ret = (char*)malloc(l+1);
  if (e == 1)
    memset(*dest,0x0,l+1);
  memcpy(*dest,buf+s,l);
  return 1;
}

/* Returns the string aligned to the left */
int pad_right(char *str,char **dest, char chr, int length)
{
  char *tempstr;
  //*dest = (char *)malloc(length+1);
  tempstr = (char *)malloc(length+1);
  memset(tempstr,chr,length - strlen(str));
  tempstr[length - strlen(str)] = '\0';
  sprintf(*dest,"%s%s",str,tempstr);
  free(tempstr);
  return 1;
}

void sig_int(int signr) {
	
	RUNNING=0;
}

void sig_chld(int signr) {
	while (waitpid(-1, NULL, WNOHANG) > 0);
}

void d_log(char *fmt, ...) {
 time_t     timenow;
 FILE       *file;
 va_list    ap;

 va_start(ap, fmt);
 timenow = time(NULL);
 if (LOG_FILE!=NULL) {
 if ((file = fopen(LOG_FILE, "a")) != NULL ) {
    fprintf(file, "%.24s - ", ctime(&timenow));
    vfprintf(file, fmt, ap);
    fclose(file);
    } else {
		printf("%.24s - ", ctime(&timenow));
		vprintf(fmt, ap);
    }} else {
		printf("%.24s - ", ctime(&timenow));
		vprintf(fmt, ap);
	}
}

char *read_field(char fname[4096], char *field) {
 char buffer[4096];
 char lfield[4096];
 char *retv=NULL;
 char *l,*r;
 FILE *f;
 if (!(f=fopen(fname,"r"))) {
     return retv;
 } else {
     while (fgets(buffer,4096,f) != NULL ) {
           sscanf(buffer,"%s",lfield);
           if (strcmp(lfield,field)==0) {
              l=malloc(4096);
              r=l;
              strcpy(l,buffer);

              l+=strlen(lfield);
              while ((l[0]==32 || l[0]==9) && l[0]!=0 ) l++;
              fclose(f);
              retv=malloc(4096);
              strcpy(retv,l);
              free(r);
              r=retv;
              while (r[0]!=13 && r[0]!=0 && r[0]!=10 && r[0]!=32 && r[0]!=9) r++;
              r[0]=0;
              return retv;
           }
     }
 }
 fclose(f);
 return retv;
}



int acceptnb(int s, struct sockaddr_in *cs, int maxsec)
{
    int flags, n;
    fd_set ac_s;
    int len;
    struct timeval tval;

    flags = fcntl(s, F_GETFL, 0);
    if (flags == -1)
        return (-1);
    n = fcntl(s, F_SETFL, flags | O_NONBLOCK);
    if (flags == -1)
        return (-1);

    FD_ZERO(&ac_s);
    FD_SET(s, &ac_s);
    tval.tv_sec = maxsec;
    tval.tv_usec = 0;

    n = select(s + 1, &ac_s, NULL, NULL, maxsec ? &tval : NULL);
    if (n == 0)
        return (0);

    if (FD_ISSET(s, &ac_s)) {
        len = sizeof(struct sockaddr_in);
        n = accept(s, (struct sockaddr *) cs, &len);
        if (n == -1) {
            switch (errno) {
            case EWOULDBLOCK:
            case ECONNABORTED:
#if defined(LINUX)
            case EPROTO:
#endif
            case EINTR:
                if (fcntl(s, F_SETFL, flags) == -1)
                    return (-1);
                return (0);
            default:
                return (-1);
            }
        }
        if (fcntl(s, F_SETFL, flags) == -1)
            return (-1);
        return (n);
    }
    if (fcntl(s, F_SETFL, flags) == -1)
        return (-1);
    return (0);
}


int check_command(char *line,FILE **file,char **CMD, char **PAYLOAD, int *payloadSize, char **CURRENT_PATH, char TARGET_PATH[512], int DEBUG)
{
  char *cmd;
  char *ptr;
  char *fileName,*tmp_dname,*tmp_fname;
  char *targetfileName;
  struct stat st;
  int size1,size2;
  DIR *dirp;
  struct dirent *direntp;
  char *ext;
  static char *tempbuf;
  int cnt;
  int ret=0;
  char TEMP_PATH[512];

  *CMD = (char *)malloc(MAX_MSG_LENGTH);
  memset(*CMD,0x0,MAX_MSG_LENGTH);
  cmd = malloc(10);

  if (!substr(line,&cmd,0,4,1))
    {
      printf("substr error\n");
      free(cmd);
      return ERROR;
    }	
  
  if (DEBUG >= 2)
    d_log("LINE:%s CMD: %s\n", line, cmd);
  
  /***************** XBOX IS CONNECTING *****************/
  if (strcmp(line,"HELLO XSTREAM 6.0") == 0)
    {
      if (DEBUG >= 1) d_log("Greetings, xbox !\n");
      strcpy(*CMD,"HELLO XBOX!");
      ret = 1;
    }

  /****************** OPEN A FILE ***********************/
  if (strcmp(cmd,"OPEN") == 0)
    {
      /* open the file */
      if (DEBUG >= 1) d_log("OPEN\n");
      
      /* Get the argument after comma */
      ptr = strtok(line,",");
      ptr = strtok(NULL,",");
      
      fileName = (char *)malloc(strlen(ptr)+1);
      targetfileName = (char *)malloc(strlen(TARGET_PATH)+strlen(ptr)+2);
      
      /* Check if the file is already opened, since the XBPlayer code does not send the CLSE command yet */
      if (*file != NULL) fclose(*file);
      
      /* Let's strip the $ character to get the true file name */
//      substr(ptr,&fileName,1,strlen(ptr)-1,1);
      strcpy(fileName, ptr);

      /* Create a new filename containing the default media path */
//      sprintf(targetfileName, "%s%s",TARGET_PATH,*CURRENT_PATH);

      /* Attach the requested filename to the end. */
	  sprintf(targetfileName,"%s%s",TARGET_PATH,fileName);
//      strcat(targetfileName, fileName);

      /* If /../ exists in our filename, we will report it and refuse */
      /* to open the requested file.  Otherwise processing should continue */
      if (!strstr(fileName,"/../")) 
	{
	  if (!(*file = fopen(targetfileName,"r")))
	    {
	      if (DEBUG >= 1)	{
			  d_log("Cannot open file: %s\n",targetfileName);
//			  perror("Cannot open file ");
		  }	      
		  sprintf(*CMD,"%-32s","-1 UNABLE TO OPEN");
	      free(fileName);
	      free(targetfileName);
	      ret = 1;
	    }
	  else
	    {
	      fstat(fileno(*file),&st);
		  /* set it to 32 char, so player understands it.*/
	      sprintf(*CMD,"%-29d OK",(int)st.st_size);
		  
	      free(fileName);
	      free(targetfileName);
	      ret = 1;
	    }
	}
      else d_log("Illegal string in filename: %s\n", fileName); 
    }

  /***************************** READ A PART OF A FILE ****************/
  if (strcmp(cmd,"READ") == 0)
    {
      /* read part of the file */
      if (!(strchr(line,',')))
	printf("No comma in command line, ignoring\n");
      else
	{
	  /* let's get the second token, it's the length of the packet wanted */
	  ptr = strtok(line,",");
	  ptr = strtok(NULL,",");
	  size1 = atoi(ptr);
	  ptr = strtok(NULL,",");
	  size2 = atoi(ptr);
	  
	  if (size2 > 0)
	    {
	      if ((file == NULL))
		{
		  d_log("File not opened !\n");
		  /* Send a 0 OK message so XBPlayer knows it's the "end" of the stream... */
		  sprintf(*CMD,"%-29d OK",0);
		  ret = 1;
		}
	      else
		{
		  *PAYLOAD = (char *)calloc(size2,1);
		  fseek(*file,size1,SEEK_SET);
		  if ((*payloadSize = fread((*PAYLOAD),1,size2,*file)))
		    {
			  /* set it to 32 char, so player understands it.*/
		      sprintf(*CMD,"%-29d OK",*payloadSize);
		      /* let's pad the result to the right of a 32 bytes string */
		      ret = 2;
		    }
		  else
		    {
		      if (feof(*file))
			{
			  /* We have reached the end of the file, send a 0 length header so the xbox knows
			     about it ! */
			  if (DEBUG >= 1)
			    d_log("Past file end !\n");
//			  strcpy(*CMD,"-1 READ FAILED");
			  sprintf(*CMD,"%-32s","-1 READ FAILED");
			  ret = 1;
			}
		      else
			d_log("can't read from file !\n");
		    }
		  
		  /* Do some cleanup ! */
		}
	    } /* buffer size is ok */
	  else
	    d_log("0 or invalid paramater for READ, ignoring\n");
	} /* comma detected */
    }

  /******************************* TELL ************************/
  /* really not sure if we still need this command             */
  /* soneone has any advice                                    */
  /* anyway changed it to 32 chars too                         */

  if (strcmp(cmd,"TELL") == 0)
    {
      /* sends the current position in the file stream */
      size1 = ftell(*file);
      if (size1 >= 0)
	{
	  sprintf(*CMD,"%-29d OK",size1);
	  ret = 1;
	}
      else
	{
	  strcpy(*CMD,"-1 TELL FAILED");
	  ret = 1;
	}
    }

  /***************************** CLOSE ************************/
  if (strcmp(cmd,"CLSE") == 0)
    {
      /* close the file */
      if (*file != NULL)
	{
	  fclose(*file);
	  d_log("Closing file\n");
	}
      else	d_log("File not opened ?\n");
      *file = NULL;
      
      /* send the answer */
      /* set it to 32 char, so player understands it.*/
      sprintf(*CMD,"%-32s","0 OK");
      ret = 1;
    }
  
  /***************************** DBUG ************************/
  /* really not sure if we still need this command             */
  /* soneone has any advice                                    */
  /* anyway changed it to 32 chars too                         */
  if (strcmp(cmd,"DBUG") == 0)
    {
//      strcpy(*CMD,"0 OK");
	  sprintf(*CMD,"%-32s","0 OK");
      ret = 1;
    }
  
  /***************************** INFO ************************/
  /* anyway changed it to 32 chars too                         */
  if (strcmp(cmd,"INFO") == 0)
    {
      /* Not implemented yet ! */
//      strcpy(*CMD,"-1 NO INFO");
	  sprintf(*CMD,"%-32s","-1 NO INFO");
      ret = 1;
    }

  /***************************** *CAT ************************/
  /* ok, we need *CURRENT_PATH here if we got the BACK cmd   */
  /* in all other cases we use full paths from now on, so    */
  /* mp3 playlists work from now on                          */
  if (strcmp(cmd,"*CAT") == 0)
    {
	  if (DEBUG >= 5) d_log("CurDirectory: %s\n",*CURRENT_PATH);

      /* Get the character after the comma, if any */
      ptr = strtok(line,",");
      if ((ptr = strtok(NULL,",")))	{
		  /* Check the command */
		  if (!strcmp(ptr,"BACK")) {
			  /* We can't go back if the current path is empty ! */
			  if (strlen(*CURRENT_PATH) > 0) {
				  tempbuf = (char *)malloc(strlen(*CURRENT_PATH)+1);
				  fileName = (char *)malloc(strlen(*CURRENT_PATH)+1);
				  strcpy(tempbuf,rindex(*CURRENT_PATH,'/'));
				  substr(*CURRENT_PATH,&fileName,0,strlen(*CURRENT_PATH)-strlen(tempbuf),1);
				  strncpy(*CURRENT_PATH,fileName,MAX_NAME_LENGTH);
				  if (DEBUG >= 10) d_log("next CurDirectory: %s\n",*CURRENT_PATH);
				  free(tempbuf);
				  free(fileName);
			  }
		  } else {
			  if (ptr[strlen(ptr)-1]=='/') ptr[strlen(ptr)-1]=0;
			  strncpy(*CURRENT_PATH,ptr,MAX_NAME_LENGTH);
		  }
	  }

	  if (DEBUG >= 5) d_log("targetpath: %s - curpath: %s - ptr: %s\n",TARGET_PATH,*CURRENT_PATH,ptr);
      sprintf(TEMP_PATH,"%s%s",TARGET_PATH,*CURRENT_PATH);
      if (DEBUG >= 1) d_log("Directory: %s\n",TEMP_PATH);

      *PAYLOAD = (char *)malloc(strlen("<SHARES>\n")+1);
      sprintf(*PAYLOAD,"<SHARES>\n");
	  cnt = 0;
	  tmp_dname=(char*)malloc(strlen(*CURRENT_PATH)+1);
	  strcpy(tmp_dname,*CURRENT_PATH); 
      if ((dirp = opendir(TEMP_PATH))!= NULL)
      while ((direntp = readdir(dirp)) != NULL)	{
		  targetfileName = (char *)malloc(strlen(TEMP_PATH)+strlen(direntp->d_name)+3);
		  sprintf(targetfileName,"%s/%s",TEMP_PATH,direntp->d_name);
		  stat(targetfileName,&st);
		  if (S_ISDIR(st.st_mode))  {
		      if (strcmp(direntp->d_name,".") && strcmp(direntp->d_name,".."))		{
				  tempbuf = (char *)malloc(strlen(tmp_dname)+strlen(direntp->d_name)+55);
				  sprintf(tempbuf,"<ITEM><ATTRIB>16</ATTRIB><PATH>%s/%s</PATH></ITEM>\n",tmp_dname,direntp->d_name);
				  if (DEBUG >= 3) d_log("-> %s/%s\n",tmp_dname,direntp->d_name);
				  *PAYLOAD = (char *)realloc(*PAYLOAD,strlen(tempbuf)+strlen(*PAYLOAD)+2);
				  strcat(*PAYLOAD,tempbuf);
				  free(tempbuf);
				  cnt++;
			  }
		  }
		  if (S_ISREG(st.st_mode)) {
	    	  /* let's get the extension */
              tmp_fname=(char *)malloc(strlen(direntp->d_name)+1);
              strcpy(tmp_fname,direntp->d_name);
		      ext = strtok(tmp_fname,".");
		      while ((ptr = strtok(NULL,"."))) strcpy(ext,ptr);
	    	  if (ext != NULL)	{
					  /* Remove the PATH and just give the server the items */
				      /* no! we send the server the full path from now on */
				  tempbuf = (char *)malloc(strlen(tmp_dname)+strlen(direntp->d_name)+55);
				  sprintf(tempbuf,"<ITEM><ATTRIB>128</ATTRIB><PATH>%s/%s</PATH></ITEM>\n",tmp_dname,direntp->d_name);
				  if (DEBUG >= 2) d_log("-> %s/%s\n",tmp_dname,direntp->d_name);
				  *PAYLOAD = (char *)realloc(*PAYLOAD,strlen(tempbuf)+strlen(*PAYLOAD)+2);
				  strcat(*PAYLOAD,tempbuf);
				  free(tempbuf);
				  cnt++;
			  }
			  free(tmp_fname);
		  }
		  free(targetfileName);
	  }
      closedir(dirp);
	  free(tmp_dname);

      if (cnt > 0)	{
		  /* We found at least one media file, we can finish the buffer*/
		  *PAYLOAD = (char *)realloc(*PAYLOAD,strlen(*PAYLOAD)+strlen("</SHARES>")+1);
		  strcat(*PAYLOAD,"</SHARES>");
		  sprintf(*CMD,"%-29d OK",strlen(*PAYLOAD));
		  *payloadSize = strlen(*PAYLOAD);
		  ret = 2;
	  } else { /* No file found ! */
		  sprintf(*CMD,"%-32s","-1 READ FAILED");
		  strcpy(*CURRENT_PATH,"");
		  ret = 1;
	  }
	}

  free(cmd);
  return ret;
}

extern unsigned long inet_addr(const char *cp);
extern char *inet_ntoa(struct in_addr in);

int main (int argc, char *argv[]) {
  int sd, newSd, cliLen;
  int ret;
  int flag;
  int opt;
  int DEBUG=0;
  char CHANGE_USER=0;
  struct in_addr INT_LISTEN_ADDR;
  int SERVER_PORT=C_SERVER_PORT;
  int nbyt;
#if defined(LINUX)
  int fd;
#endif
  char *line;
  char *msg;
  char *cmd;
  char *tmp=NULL;
  char *config_file=NULL;
  char *linetmp;
  char *CMD;
  char *PAYLOAD;
  char *CURRENT_PATH;
  caddr_t adr;
  int payloadSize;
  char user[16]=USER_NAME,groupname[16]=GROUP_NAME; /* yeah too big...sue me */
  char parameter[16]="";
  struct passwd *u_pwd;
  struct group *g_group;
  char TARGET_PATH[512];
  struct sockaddr_in cliAddr, servAddr;
  FILE *file = NULL, *pid_file;
  /* Need to set this up early. */
  strcpy(TARGET_PATH, MEDIA_FILES_PATH);
  INT_LISTEN_ADDR.s_addr=(unsigned long)inet_addr(LISTEN_ADDR);
	  
  while ((opt = getopt(argc, argv, "i:d:u:g:p:D:c:l:h")) != EOF)
      switch(opt) {
        case 'c':
			config_file=(char *)malloc(strlen(optarg)+1);
			strcpy(config_file,optarg);
			break;
        case 'i':
			strcat(parameter,"i");
            INT_LISTEN_ADDR.s_addr=inet_addr(optarg);
            break;
        case 'd':
			strcat(parameter,"d");
            strncpy(TARGET_PATH,optarg,511);
            break;
        case 'u':
			strcat(parameter,"u");
            strncpy(user,optarg,15);
            CHANGE_USER=1;
            break;
        case 'g':
			strcat(parameter,"g");
            strncpy(groupname,optarg,15);
            break;
        case 'p':
			strcat(parameter,"p");
            SERVER_PORT=atoi(optarg);
            break;
        case 'D':
			strcat(parameter,"D");
            DEBUG=atoi(optarg);
            break;
        case 'h':
            usage(argv[0]);
            break;
        case 'l':
	        LOG_FILE=(char *)malloc(strlen(optarg)+1);
	        strcpy(LOG_FILE,optarg);
            break;
        case '?':
            printf("Unknown option: %s\n", optarg);
            usage(argv[0]);
            break;
        default:
            usage(argv[0]);
            break;
	  }
  if (config_file==NULL) {
	  config_file=(char *)malloc(strlen(CONFIG)+1);
	  strcpy(config_file,CONFIG);
  }
  tmp=read_field(config_file,"USER");
  if (tmp!=NULL) {
	  if (strchr(parameter,'u')==NULL) strncpy(user,tmp,15);
	  free(tmp);
  }
  tmp=read_field(config_file,"GROUP");
  if (tmp!=NULL) {
      if (strchr(parameter,'g')==NULL) strncpy(groupname,tmp,15);
      free(tmp);
  }
  tmp=read_field(config_file,"PORT");
  if (tmp!=NULL) {
	  if (strchr(parameter,'p')==NULL) SERVER_PORT=atoi(tmp);
	  free(tmp);
  }
  tmp=read_field(config_file,"LOCAL_IP");
  if (tmp!=NULL) {
	  if (strchr(parameter,'i')==NULL) INT_LISTEN_ADDR.s_addr=inet_addr(tmp);
	  free(tmp);
  }
  tmp=read_field(config_file,"PATH");	
  if (tmp!=NULL) {
	  if (strchr(parameter,'d')==NULL) strncpy(TARGET_PATH,tmp,511);
	  free(tmp);
  }
  tmp=read_field(config_file,"LOGFILE");
  if (tmp!=NULL) {
	  if (strchr(parameter,'l')==NULL) {
		  LOG_FILE=(char *)malloc(strlen(tmp)+1);
		  strcpy(LOG_FILE,tmp);
	  }
	  free(tmp);
  }

  if (user[0]!=0 && (getuid() == 0 || CHANGE_USER)) {
	  /* running as root or got option by cmd line, try to change user */
	  u_pwd=getpwnam(user);
	  if (u_pwd==NULL) {
		  d_log("ERROR! Could not get uid for user %s\n",user);
		  return -1;
	  }
	  if (setuid(u_pwd->pw_uid)!=0) {
		d_log("WARNING! Could not change to user %s(%i): %s\n",user,u_pwd->pw_uid,strerror(errno));
	  } else 
		  d_log("Changed to user: %s(%i)\n",user,u_pwd->pw_uid);
	  if (groupname[0]!=0) {
	      g_group=getgrnam(groupname);
    	  if (g_group==NULL) {
	          d_log("ERROR! Could not get gid for group %s\n",groupname);
    	      return -1;
	      }
    	  if (setgid(g_group->gr_gid)!=0) {
        	d_log("WARNING! Could not change to group %s(%i): %s\n",groupname,g_group->gr_gid,strerror(errno));
	      } else
    	      d_log("Changed to group: %s(%i)\n",groupname,g_group->gr_gid);
	  }
  }
//  CURRENT_PATH = (char*)malloc(MAX_MSG_LENGTH);
#if defined(LINUX)
  if ( (fd=open("/dev/zero",O_RDWR))<0) {
	  perror("cannot open /dev/zero");
      return ERROR;
  }
  if ((adr = mmap(0,MAX_NAME_LENGTH ,PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0)) == (caddr_t)-1) {
	  perror("cannot open shared mem");
	  return -1;
  }
  CURRENT_PATH = (char *)adr;
  close(fd);
#else
  if ((adr = mmap(0,MAX_NAME_LENGTH ,PROT_READ | PROT_WRITE, MAP_ANON | MAP_SHARED, -1, 0)) == (caddr_t)-1) {
	  perror("cannot open shared mem");
	  return -1;
  }
  CURRENT_PATH = (char *)adr;
#endif

		

	  
  cmd = (char*)malloc(MAX_CMD_LENGTH);
  linetmp = (char*)malloc(MAX_MSG_LENGTH);
  line = (char*)malloc(MAX_MSG_LENGTH);
  msg = (char*)malloc(MAX_MSG_LENGTH);

  strcpy(CURRENT_PATH,"");
  d_log("Address: %s\n",inet_ntoa((struct in_addr)INT_LISTEN_ADDR));
  d_log("Port: %u\n", SERVER_PORT);
  d_log("Directory: %s\n", TARGET_PATH);

  /* create socket */
  sd = socket(AF_INET, SOCK_STREAM, 0);
  if (sd < 0)
    {
      perror("cannot open socket ");
      return ERROR;
    }
  
  /* Does not seem to work on linux 2.4.18+ */
  /* at least SO_REUSEADDR does work fine with 2.4.19+ */
  flag = 1;
  setsockopt (sd, SOL_SOCKET, SO_REUSEADDR, (char *) &flag, sizeof(int));
  setsockopt (sd, IPPROTO_TCP, TCP_NODELAY, (char *) &flag, sizeof(int));

  /* bind server port */
  servAddr.sin_family = AF_INET;
  servAddr.sin_addr.s_addr = INT_LISTEN_ADDR.s_addr;
  servAddr.sin_port = htons(SERVER_PORT);
  if (bind(sd, (struct sockaddr *) &servAddr, sizeof(servAddr)) < 0)
    {
      perror("cannot bind port ");
      return ERROR;
    }
  listen(sd,5);
  if ((nbyt = fork()) != -1) 
	  if (nbyt == 0) {
       	if (!(pid_file = fopen(PIDFILE, "w"))) {
		     fprintf(stderr, "cannot create pidfile %s: %s\n", PIDFILE, strerror(errno));
		} else {
			fprintf(pid_file,"%i",getpid());
			fclose(pid_file);
		}
	  signal(SIGCHLD,sig_chld); 
      signal(SIGINT,sig_int);
      signal(SIGTERM,sig_int);
						
	  
  if (DEBUG >= 1) d_log("%s: waiting for commands on port TCP %u\n",argv[0],SERVER_PORT);
  
  while (RUNNING) {

      cliLen = sizeof(cliAddr);
      /* Create the client socket */
      newSd = acceptnb(sd, (struct sockaddr_in *) &cliAddr, 2);
//      newSd = accept(sd, (struct sockaddr *) &cliAddr, &cliLen);
      if (newSd == -1) {
		  continue;
	  }

      flag = 1;
      setsockopt (newSd, IPPROTO_TCP, TCP_NODELAY, (char *) &flag, sizeof(int));
      flag = 2048;
      setsockopt (newSd, IPPROTO_TCP, TCP_MAXSEG, (char *) &flag, sizeof(int));

	  if (newSd == 0) {
		  //we handle the finished child processes...
		  while (waitpid(-1, NULL, WNOHANG) > 0);
		  continue;
	  }
	  
      if ((nbyt = fork()) == -1) {
		  fprintf(stdout, "fork: %s\n", strerror(errno));
		  shutdown(newSd, 2);
		  close(newSd);
	  }
	  if (nbyt == 0) {
	      /* init line */
    	  memset(line,0x0,MAX_MSG_LENGTH);

	      /* receive commands */
    	  while(read(newSd,line,MAX_MSG_LENGTH) > 0) {
	          /* strip all unwanted characters */
    	      if ((strchr(line,10))) {
        	      substr(line,&linetmp,0,strlen(line)-1,1);
            	  strcpy(line,linetmp);
	          }
    	      if ((strchr(line,13)))   {
        	      substr(line,&linetmp,0,strlen(line)-1,1);
            	  strcpy(line,linetmp);
	          }
	
    	      /* Check what do do with the line */
        	  ret = check_command(line,&file,&CMD,&PAYLOAD,&payloadSize,&CURRENT_PATH, TARGET_PATH, DEBUG);
	          /* There is a CMD */
    	      if (ret >= 1) {
        	      pad_right(CMD,&linetmp,' ',strlen(CMD));
            	  strcpy(msg,linetmp);
	              write(newSd,msg,strlen(msg));
    	          if (DEBUG >= 1)
        	          d_log("%i -> %s\n",strlen(msg),msg);
            	  free(CMD);
	          }
    	      /* There is a payload too */
        	  if (ret == 2) {
            	  write(newSd,PAYLOAD,payloadSize);
	              free(PAYLOAD);
    	      }
        	  memset(line,0x0,MAX_MSG_LENGTH);
	      } /* while(read) */

	      close(newSd);
    	  file = NULL;
	      if (DEBUG >= 1) d_log("Connection closed\n");
	      free(linetmp);
  		  free(line);
		  free(msg);
		  free(cmd);
		  free(LOG_FILE);
		  return 0;
	  } /* fork child end (nbyt == 0)*/

//	  close(newSd);
	  file = NULL;
	  
    } /* while (1) */	  
    d_log("closing xbms\n");
	unlink(PIDFILE);
   }
  free(linetmp);
  free(line);
  free(msg);
  free(cmd);
  free(LOG_FILE);
  return 0;
}


